#!/usr/bin/env python3
#
# Capture some statististics about umaudemc pcheck using the Python profiler
#

import csv
from datetime import datetime
import os
import shlex
import shutil
import subprocess
import sys


def comparison(table):
	"""Comparison of some values"""

	quo = table.external / table.total

	print(f'maude.init()        {table.maude_init.mean():.5f} {table.maude_init.std():.5f}')
	print(f'maude.load()        {table.maude_load.mean():.5f} {table.maude_load.std():.5f}')
	print(f'graph               {table.graph.mean():.5f} {table.graph.std():.5f}')
	print(f'total vs external   {table.total.corr(table.external)}')
	print(f'total vs graph      {table.total.corr(table.graph)}')
	print(f'external (%)        {quo.min() * 100:.2f}-{quo.max() * 100:.2f}')


def header(text):
	"""Print a section header"""

	print(f'\n\x1b[33m# {text}\x1b[0m\n')


def extract_bare_command(cmdline):
	"""Common part of a pcheck command invokation"""

	return cmdline.split(' -- ', maxsplit=1)[0].replace(' --backend storm', '')


def extract_backend(row):
	"""Deduce the backend used for the given row of the results table"""

	if '--backend storm' in row.command:
		return 'storm' if row.storm_cmd else 'stormpy'
	else:
		return 'ngprism' if row.ngprism else 'prism'


def backend_summary(group, pd=None):
	"""Make a row with the total time per backend using groupby"""

	if len(group.backend) < 4:
		print(f'\x1b[93mWarning:\x1b[0m case not covered by all backends (only {" and ".join(group.backend)}):')
		print(f'  {list(group.bare_command)[0]}')
		return None

	values, columns = [], []

	for _, row in group.iterrows():
		backend = row.backend

		# Total execution time
		values.append(row.total)
		columns.append(('total', backend))
		# pcheck execution time
		values.append(row.pcheck)
		columns.append(('pcheck', backend))
		# Model load time
		values.append(row.maude_load)
		columns.append(('load', backend))
		# Graph generation time
		values.append(row.graph)
		columns.append(('graph', backend))
		# Expansion of graph to file
		values.append(row.prism_graph)
		columns.append(('prism_graph', backend))
		# External time
		if backend == 'stormpy':
			values.append(row.stormpy_model_checking + row.stormpy_build_model)
		else:
			values.append(row.external)
		columns.append(('external', backend))

	return pd.Series(values, pd.MultiIndex.from_tuples(columns)).sort_index()


def analyze(filename):
	import pandas as pd
	import matplotlib.pyplot as plt

	res = pd.read_csv(filename)

	# Get the bare command without backend-specific additions
	res['bare_command'] = res.command.apply(extract_bare_command)
	# Make the backend information explicit
	res['backend'] = res.apply(extract_backend, axis=1)

	summary = res.groupby('bare_command').apply(backend_summary, pd=pd).dropna()
	summary.to_html('summary4case.htm')

	# Show the total time per backend and the ratios
	summary_sum = summary.sum()
	fastest = min(summary_sum.total)

	header('Total execution time per backend')

	for backend, time in summary_sum.total.items():
		print(f'{backend:10}   {time:10.2f}    {time / fastest:10.4f}')

	header('Execution time percentage per backend')

	print(f'{"":10} {"external":>10} {"graph":>10} {"load":>10} {"write":>10}')

	for backend in summary_sum.total.keys():
		r_external = 100 * summary_sum.external[backend] / summary_sum.pcheck[backend]
		r_graph = 100 * summary_sum.graph[backend] / summary_sum.pcheck[backend]
		r_load = 100 * summary_sum.load[backend] / summary_sum.pcheck[backend]
		r_write = 100 * summary_sum.prism_graph[backend] / summary_sum.pcheck[backend]

		print(f'{backend:10} {r_external:10.2f} {r_graph:10.2f} {r_load:10.2f} {r_write:10.2f}')

	# Other calculations
	# res.sort_values('total', inplace=True)
	storm_filter = res.command.str.contains('--backend storm')
	ngprism = res[res.ngprism]
	prism = res[~ storm_filter & ~ res.ngprism]
	storm = res[storm_filter]
	stormpy = res[storm_filter & ~ res.storm_cmd]

	header('All cases with ngPRISM')
	comparison(ngprism)

	header('Methods other than strategy with ngPRISM')
	comparison(ngprism[~ ngprism.command.str.contains('--assign strategy')])

	header('Strategy method with ngPRISM')
	comparison(ngprism[ngprism.command.str.contains('--assign strategy')])

	header('All cases with Storm')
	comparison(storm)

	header('All cases with StormPy')
	comparison(stormpy)

	header('All cases with PRISM')
	comparison(prism)

	quo = res.external / res.total
	plt.plot(res.total, quo, label='External (%)')
	plt.plot(res.total, res.graph / res.total, label='Graph (%)')
	plt.legend(loc='best')
	plt.show()


if __name__ == '__main__':
	import argparse

	parser = argparse.ArgumentParser(description='Profiler for umaudemc\'s pcheck')
	parser.add_argument('--analyze', help='Analyze the given CSV file')
	parser.add_argument('--ngprism', help='Run PRISM examples with ngprism too (you should run prism -ng apart)', action='store_true')
	parser.add_argument('--stormpy', help='Run Storm examples with and without StormPy', action='store_true')

	args = parser.parse_args()

	if args.analyze:
		analyze(args.analyze)
		sys.exit(0)

	# Auxiliary script for profiling a given test case
	profiler_case = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'profiler-case.py')

	# Test cases are pcheck commands in the profiler.txt file
	with open('profiler.txt') as inputfile:
		csv_name = f'profile-{datetime.now().strftime("%Y%m%d-%H%M%S")}.csv'

		# Write the header of the CSV file, the data rows will be written by profiler-case
		with open(csv_name, 'w') as csvfile:
			csvw = csv.writer(csvfile)
			csvw.writerow(['command', 'ngprism', 'storm_cmd', 'total', 'pcheck', 'external', 'graph', 'parse_input', 'maude_init', 'maude_load', 'prism_graph', 'stormpy_build_model', 'stormpy_model_checking'])

		for line in inputfile:
			# Run the example with PRISM or Storm (depending on its --backend argument)
			subprocess.run([sys.executable, profiler_case, csv_name] + shlex.split(line))

			# If we want test PRISM under NailGun
			if args.ngprism and '--backend storm' not in line:
				mod_env = os.environ.copy()
				mod_env['PRISM_PATH'] = shutil.which('ngprism')
				subprocess.run([sys.executable, profiler_case, csv_name] + shlex.split(line), env=mod_env)

			# If we want to test Storm without StormPy
			if args.stormpy and '--backend storm' in line:
				mod_env = os.environ.copy()
				mod_env['NO_STORMPY'] = 'true'
				subprocess.run([sys.executable, profiler_case, csv_name] + shlex.split(line), env=mod_env)
